import pytest
from ros_get.__main__ import parse_args


def test_parse_no_command(capsys):
    """It should print an error message when no command is provided"""
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        parse_args([])
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code != 0

    captured = capsys.readouterr()
    assert captured.out == ''
    assert 'the following arguments are required' in captured.err or 'too few arguments' in captured.err


def test_parse_status():
    func, args = parse_args(['status'])
    assert func == 'status'
    assert not args.verbose


def test_parse_status_verbose():
    func, args = parse_args(['--verbose', 'status'])
    assert func == 'status'
    assert args.verbose



from .commands import install, update, status, list_packages, remove
from .workspace import create, switch, save, list_workspaces, locate, name, rosdistro_url

__version__ = '0.4.0'
__url__ = 'https://github.com/Rayman/ros-get'
__author__ = 'Ramon Wijnands'
__email__ = 'rayman747@hotmail.com'

__all__ = [
    'install', 'update', 'status', 'list_packages', 'remove', 'create', 'switch', 'save', 'list_workspaces', 'locate',
    'name', 'rosdistro_url'
]



import imp

import pytest
import ros_get.workspace
import xdg
from ros_get import create, install, list_workspaces, locate, remove, save, switch, update


@pytest.fixture()
def empty_config_home(tmpdir, monkeypatch):
    """
    Monkeypatch XDG_CONFIG_HOME to an empty directory in /tmp
    """
    monkeypatch.setenv('XDG_CONFIG_HOME', str(tmpdir))
    imp.reload(xdg)
    imp.reload(ros_get.workspace)
    return tmpdir


def test_fixture(empty_config_home):
    assert ros_get.workspace.config_dir.startswith('/tmp')


def test_install(empty_config_home):
    assert install([], verbose=True) == 1


@pytest.mark.skip()
def test_update():
    assert update(restore_versions=False, verbose=True) == 1


def test_list(empty_config_home):
    list_workspaces(verbose=True)


def test_remove(empty_config_home):
    remove(['unknown'], verbose=True)


@pytest.mark.skip()
def test_ws_create(empty_config_home):
    # TODO: implement when the final implementation of create is finished
    create(verbose=True)


def test_ws_switch(empty_config_home):
    switch('unknown', verbose=True)


def test_ws_save(empty_config_home):
    save('dir', 'name', verbose=True)


def test_ws_list(empty_config_home):
    assert len(empty_config_home.listdir()) == 0

    list_workspaces(verbose=True)

    config = empty_config_home.join('ros-get')
    assert config.check(dir=1)
    assert config.join('workspace').check(exists=0)
    assert config.join('workspaces').check(dir=1)


def test_ws_locate(empty_config_home):
    locate(verbose=True)
    assert len(empty_config_home.listdir()) == 0



import logging
import os
import shlex
import subprocess
import sys
from argparse import Namespace

import xdg
from catkin_tools.metadata import find_enclosing_workspace
from catkin_tools.terminal_color import ColorMapper
from catkin_tools.verbs import catkin_config, catkin_build
from future.moves.urllib.error import URLError
from future.moves.urllib.request import urlopen

from .utils import mkdir_p, symlink_force, __name__ as utilsname

logger = logging.getLogger(__name__)

config_dir = os.path.join(xdg.XDG_CONFIG_HOME, 'ros-get')
ws_file = os.path.join(config_dir, 'workspace')
ws_dir = os.path.join(config_dir, 'workspaces')

clr = ColorMapper().clr


def create(rosdistro_index_url, extend_path, dir, name, build_tool, verbose):
    """Creates a new workspace, saves it, and switches to it if it is the first
    workspace.

    :param rosdistro_index_url: The rosdistro to use
    :param extend_path: Parent workspace to use.
    :param dir: Where to create the workspace
    :param name: Name of the new workspace.
    :param name: Create the workspace with colcon, instead of catkin_tools
    :param verbose: Unused.
    """
    # also allow files
    if os.path.isfile(rosdistro_index_url):
        rosdistro_index_url = 'file://%s' % os.path.realpath(rosdistro_index_url)

    try:
        urlopen(rosdistro_index_url)
    except (ValueError, URLError) as e:
        logger.error(e)
        return 1

    if not os.path.isdir(dir):
        logger.error('target path is not a directory')
        return 1

    enclosing_workspace = find_enclosing_workspace(dir)
    if enclosing_workspace:
        logger.error("Its not allowed to create a worksapce inside another workspace, other workspace found here:\n%s",
                     enclosing_workspace)
        return 1

    # try to guess which build tool to use
    if os.path.exists(os.path.join(extend_path, '.catkin')):
        build_tool = 'catkin_tools'
    else:
        build_tool = 'colcon'

    if build_tool == 'catkin_tools':
        result = create_workspace_with_catkin_tools(extend_path, dir)
    elif build_tool == 'colcon':
        result = create_workspace_with_colcon(extend_path, dir)

    if result:
        return result

    save(dir, name, verbose)
    save_config(dir, rosdistro_index_url=rosdistro_index_url)


def switch(name, verbose):
    """Switch workspace.

    :param name: Name of the workspace to switch to.
    :param verbose: Unused.
    """
    mkdir_p(ws_dir)
    dir = os.path.join(ws_dir, name)
    if not os.path.isdir(dir):
        logger.error('workspace does not exists: %s', name)
        return 1

    logging.getLogger(utilsname).setLevel(logging.ERROR)
    symlink_force(os.path.join('workspaces', name), ws_file)


def save(dir, name, verbose):
    """Internal ws-save command to register an existing workspace.

    If there is no workspace currently, the command also switches to it.

    :param name: Name of the workspace to save.
    :param verbose: Unused.
    """
    if not os.path.isdir(dir):
        logger.error('target path is not a directory')
        return 1

    abs_dir = os.path.abspath(dir)
    if not name:
        name = os.path.basename(abs_dir)

    if not name:
        logger.error('name not valid: %s', name)
        return 1

    # save the result
    mkdir_p(ws_dir)
    symlink_force(abs_dir, os.path.join(ws_dir, name))
    logger.info("saved the workspace as '%s'", name)

    if not os.path.islink(ws_file):
        logger.info("created first workspace, let's make it the default")
        switch(name, verbose)


def list_workspaces(verbose):
    """List the available workspaces.

    :param verbose: Unused.
    """
    try:
        active = os.path.relpath(os.readlink(ws_file), 'workspaces')
    except OSError:
        active = None

    mkdir_p(ws_dir)
    for link_name in os.listdir(ws_dir):
        try:
            source = os.readlink(os.path.join(ws_dir, link_name))
        except OSError as e:
            logger.error('%s => %s', link_name, repr(e))
        else:
            s = '@pf-@| @{cf}%s@| => @{cf}%s@|' % (link_name, source)
            if link_name == active:
                s += ' (@{yf}active@|)'
            if not os.path.isdir(source):
                s += ' @{rf}No such directory@|'
            print(clr(s))


def locate(verbose):
    """Print location of the current workspace.

    :param verbose: Unused.
    """
    if not os.path.islink(ws_file):
        print('no current workspace found, see "ros-get ws-create --help" how to create one')
        return 1

    else:
        print(os.path.realpath(ws_file))


def name(verbose):
    """Print name of the current workspace.

    :param verbose: Unused.
    """
    if not os.path.islink(ws_file):
        print('no current workspace found, see "ros-get ws-create --help" how to create one')
        return 1

    else:
        print(os.path.relpath(os.readlink(ws_file), 'workspaces'))


def rosdistro_url(verbose):
    ws_file = os.path.join(config_dir, 'workspace', '.ros-get')
    print(load_config(ws_file, 'rosdistro_index_url'))


def save_config(dir, **kwargs):
    ws_config_dir = os.path.join(dir, '.ros-get')
    mkdir_p(ws_config_dir)

    for k, v in kwargs.items():
        with open(os.path.join(ws_config_dir, k), 'w') as f:
            f.write(v)


def load_config(dir, key):
    ws_config_dir = os.path.join(dir, '.ros-get')
    mkdir_p(ws_config_dir)

    with open(os.path.join(dir, key)) as f:
        return f.read()


def create_workspace_with_catkin_tools(extend_path, dir):
    try:
        if catkin_config.main(
            Namespace(
                workspace=dir,
                extend_path=extend_path,
                profile=None,
                append_args=None,
                remove_args=None,
                init=True,
                mkdirs=True)):
            logger.error('catkin config error\n\nAre you sure the workspace you are extending is generated by catkin? '
                         'For colcon, pass the --colcon argument')
            return 1
    except ValueError as e:
        logger.error(e)
        return 1

    if catkin_build.main(
        Namespace(
            workspace=dir,
            develdebug=None,
            force_color=None,
            no_color=None,
            build_this=None,
            start_with_this=None,
            no_deps=None,
            profile=None,
            mem_limit=None,
            dry_run=None,
            get_env=None,
            save_config=None,
            parallel_jobs=None,
            verbose=None,
            packages=None,
            start_with=None,
            unbuilt=None,
            force_cmake=None,
            pre_clean=None,
            interleave_output=None,
            no_status=None,
            limit_status_rate=10.0,
            no_install_lock=None,
            no_notify=None,
            continue_on_failure=None,
            summarize=None)):
        logger.error('catkin build error')
        return 1


def create_workspace_with_colcon(extend_path, dir):
    cmd = '. %s && colcon build' % shlex.quote(os.path.join(extend_path, 'setup.sh'))
    logger.info('run: "%s"', cmd)
    result = subprocess.call(cmd, shell=True, cwd=dir, env={})
    if not result:
        return result

    # command failed, let's check if the reason
    if subprocess.call('command -v colcon', shell=True):
        # in Python3 we should have colcon, so the failure was due to colcon's failure
        if sys.version_info > (3,):
            return result
        else:
            # in Python2 we can't guarantee a colcon installation, the user has to provide it
            logger.error("colcon was not found. For Python2 ros-get can't add it as a dependency, so you should "
                         "install it by yourself")
            return result
    else:
        return result



#!/usr/bin/env python
import os
import re
from setuptools import setup, find_packages


def get_ini_variable(name):
    with open(os.path.join(os.path.dirname(__file__), 'src', 'ros_get', '__init__.py')) as f:
        return re.compile(r".*%s = '(.*?)'" % name, re.S).match(f.read()).group(1)


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as r_file:
    readme = r_file.read()

setup(
    name='ros_get',
    license='MIT',
    version=get_ini_variable('__version__'),
    url=get_ini_variable('__url__'),
    author=get_ini_variable('__author__'),
    author_email=get_ini_variable('__email__'),
    description='Simple tools for working with ROS source packages',
    long_description=readme,
    package_dir={'': 'src'},  # tell distutils packages are under src
    packages=find_packages('src'),  # include all packages under src
    install_requires=[
        "mock < 4; python_version < '3'",
        'argcomplete',
        'catkin_pkg',
        'catkin_tools',
        "colcon-common-extensions; python_version >= '3'",
        'colorlog',
        'future',
        'rosdep',
        'rosdistro >= 0.7.3',
        'rosinstall_generator',
        'six>=1.7',  # https://github.com/testing-cabal/mock/issues/257
        'trollius',  # remove when catkin>0.4.4 is released
        'vcstools',
        'xdg==1.0.7',
    ],
    entry_points={'console_scripts': ['ros-get=ros_get.__main__:main']}, )



import errno
import logging
import os
from argparse import Namespace
from queue import Queue, Empty

from catkin_pkg.packages import find_packages_allowing_duplicates
from rosdep2.main import command_update
from rosinstall_generator.generator import generate_rosinstall_for_repos
from vcstools import get_vcs_client

from .utils import mkdir_p, get_rosdistro, update_folder, symlink_force, rosdep_install
from .workspace import ws_file

logger = logging.getLogger(__name__)

installed_dir = os.path.realpath(os.path.join(ws_file, '.env', 'installed'))
target_path = os.path.realpath(os.path.join(ws_file, 'repos'))
link_dir = os.path.realpath(os.path.join(ws_file, 'src'))


def install(pkgs, verbose):
    pkgs_done = recursive_update(pkgs, False, verbose)
    pkgs_succeeded = [pkg for pkg in pkgs if pkg in pkgs_done]
    pkgs_skipped = set(pkgs) - pkgs_done

    for pkg in pkgs_skipped:
        logger.error("Package %s was not defined in the ros distribution", pkg)
    add_pkgs_to_installed_list(pkgs_succeeded)

    if not pkgs_done:
        return 1
    result = rosdep_install(link_dir)
    if pkgs_skipped:
        return 1
    return result


def update(restore_versions, verbose):
    # first check if a custom rosdistro has been configured
    get_rosdistro()

    logger.info('rosdep update')
    exit_code = command_update(Namespace(include_eol_distros=False, ros_distro=None))
    if exit_code:
        logger.error('`rosdep update` exited with %d', exit_code)
        return exit_code

    pkgs = get_pkgs_from_installed_list()
    pkgs_done = recursive_update(pkgs, restore_versions, verbose)

    cleanup_symlinks(pkgs_done)

    exit_code = rosdep_install(link_dir)
    if exit_code:
        logger.error('`rosdep install` exited with %d', exit_code)
        return exit_code


def status(verbose):
    distro = get_rosdistro()

    repositories = [
        r for r in distro.repositories.values() if r.source_repository and r.source_repository.patched_packages
    ]

    for path in os.listdir(target_path):
        name = os.path.basename(path)

        repo = [r for r in repositories if r.name == name]

        if len(repo) != 1:
            logger.warning('skipping unknown repo: %s', name)
            continue
        repo = repo[0]

        config = generate_rosinstall_for_repos({'DOESNOTMATTER': repo}, version_tag=False, tar=False)[0]

        assert len(config) == 1
        repo_type, attributes = next(iter(config.items()))

        try:
            url = attributes['uri']
            if 'version' in attributes:
                version = attributes['version']
        except AttributeError as e:
            logger.warning("Repository '%s' does not provide the necessary " 'information: %s' % (path, e))
            continue

        client = get_vcs_client(repo_type, os.path.join(target_path, path))

        if client.get_url() != url:
            logger.error("local vcs url is different from the distro's url:\n\tlocal:  %s\n\tdistro: %s",
                         client.get_url(), url)

        current_version = client.get_current_version_label()
        if current_version == version:
            print('=== %s (%s) ===' % (name, repo_type))
        else:
            print('=== %s (%s) === @ %s' % (name, repo_type, current_version))

        print(client.get_status(untracked=True))


def list_packages(installed, verbose):
    if installed:
        for pkg in sorted(get_pkgs_from_installed_list()):
            print(pkg)
    else:
        distro = get_rosdistro()

        pkgs = []
        for r in distro.repositories.values():
            if r.source_repository and r.source_repository.patched_packages:
                for p in r.source_repository.patched_packages:
                    pkgs.append(p)

        for p in sorted(pkgs):
            print(p)


def remove(pkgs, verbose):
    remove_pkgs_from_installed_list(pkgs)
    logger.warn("marked %d package(s) for removal", len(pkgs))
    logger.warn("please run the following command to update your workspace:\n\n\tros-get update")


def recursive_update(pkgs, restore_versions, verbose):
    if not pkgs:
        logger.warn('no package specified')
        return set()

    mkdir_p(target_path)
    # create an ignore file so that colcon won't process the repos directory
    open(os.path.join(target_path, 'COLCON_IGNORE'), 'w')

    distro = get_rosdistro()

    repositories = [
        r for r in distro.repositories.values() if r.source_repository and r.source_repository.patched_packages
    ]

    pkgs_queue = Queue()
    pkgs_done = set()
    for pkg in pkgs:
        pkgs_queue.put_nowait(pkg)

    # mapping from manifest to found location
    pkgs_manifests = {}

    try:
        while True:
            pkg = pkgs_queue.get_nowait()

            if pkg in pkgs_done:
                continue

            repo = [repo for repo in repositories if pkg in repo.source_repository.patched_packages]
            if len(repo) == 0:
                logger.debug("Package '%s' can't be found in a repository", pkg)
                continue

            assert len(repo) == 1, "Package '%s' is in multiple repositories" % pkg
            repo = repo[0]

            logger.info('installing: %s', pkg)
            if not len([m for m in pkgs_manifests if m.name == pkg]):
                update_folder(target_path, {repo.name: repo}, restore_versions, verbose)

                # which packages did we download?
                updated_packages = find_packages_allowing_duplicates(os.path.join(target_path, repo.name))

                # first check for expected packages that were not found
                found_names = set(package.name for package in updated_packages.values())
                for package in repo.source_repository.patched_packages:
                    if package not in found_names:
                        logger.warning("Package '%s' not found in the repo: '%s'", package, repo.name)

                # then check for found packages that were not in the yaml
                for subfolder, package in updated_packages.items():
                    if package.name in repo.source_repository.patched_packages:
                        logger.info("found '%s'" % os.path.join(repo.name, subfolder))
                        pkgs_manifests[package] = os.path.join(repo.name, subfolder)
                    else:
                        logger.debug("Found package '%s' in an unexpected repo: '%s'", package.name, repo.name)

            manifest = [m for m in pkgs_manifests if m.name == pkg]
            if not len(manifest):
                logger.error("Required package '%s' not found", pkg)
                exit(1)
            assert len(manifest) == 1, "Package '%s' was found multiple times" % pkg
            manifest = manifest[0]

            deps = manifest.buildtool_depends + manifest.build_depends + manifest.run_depends + manifest.test_depends
            for dep in deps:
                if dep not in pkgs_done:
                    logger.debug("queing: '%s'", dep.name)
                    pkgs_queue.put_nowait(dep.name)

            pkgs_done.add(pkg)

    except Empty:
        pass

    if not pkgs_done:
        logger.error('no repository updated, package could not be found')

    # create the symlinks in the src folder
    mkdir_p(link_dir)
    for pkg in pkgs_done:
        manifest = [m for m in pkgs_manifests if m.name == pkg]
        assert len(manifest) == 1
        manifest = manifest[0]

        folder = pkgs_manifests[manifest]

        source = os.path.join(target_path, folder)
        link_name = os.path.join(link_dir, manifest.name)

        source = os.path.relpath(source, link_dir)
        symlink_force(source, link_name)

    return pkgs_done


def cleanup_symlinks(pkgs_done):
    """Remove all symlinks from the link_dir that are not in pkgs_done"""
    logger.info("Cleaning up old symlinks")
    for f in os.listdir(link_dir):
        if f not in pkgs_done:
            filename = os.path.join(link_dir, f)
            if os.path.islink(filename):
                logger.info("removing symlink: %s", f)
                os.remove(filename)


def add_pkgs_to_installed_list(pkgs):
    mkdir_p(installed_dir)

    # touch the file
    for pkg in pkgs:
        logger.debug('marking for installation: %s', pkg)
        open(os.path.join(installed_dir, pkg), 'a').close()


def get_pkgs_from_installed_list():
    mkdir_p(installed_dir)
    return os.listdir(installed_dir)


def remove_pkgs_from_installed_list(pkgs):
    for pkg in pkgs:
        try:
            os.remove(os.path.join(installed_dir, pkg))
        except OSError as e:
            if e.errno == errno.ENOENT:
                logger.error('unable to locate package: %s', pkg)
            else:
                raise
        else:
            logger.info('Removing: %s', pkg)



import errno
import logging
import os

from rosdep2.main import _rosdep_main

from rosdistro import get_index, get_index_url, repository, get_distribution, DEFAULT_INDEX_URL
from rosdistro.source_repository_specification import SourceRepositorySpecification

from rosdep2 import RosdepLookup, create_default_installer_context, get_default_installer
from rosdep2.rospkg_loader import DEFAULT_VIEW_KEY
from rosinstall_generator.generator import generate_rosinstall_for_repos
from vcstools import get_vcs_client

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

logger = logging.getLogger(__name__)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def symlink_force(source, link_name):
    logging.info("symlink '%s' => '%s'", source, link_name)
    try:
        os.symlink(source, link_name)
    except OSError as e:
        if e.errno == errno.EEXIST:
            logger.debug('symlink already exists: %s', link_name)
            logger.debug("replacing symlink from '%s' to '%s'", os.path.realpath(link_name), source)
            os.remove(link_name)
            os.symlink(source, link_name)
        else:
            raise e


class SourceRepositorySpecificationMock(SourceRepositorySpecification):
    def __init__(self, name, data):
        super(SourceRepositorySpecificationMock, self).__init__(name, data)
        self.patched_packages = data.get('packages', [])


def get_rosdistro():
    if 'ROS_DISTRO' in os.environ:
        distroname = os.environ['ROS_DISTRO']
    else:
        raise AssertionError('ROS_DISTRO is not defined in environment')

    index_url = get_index_url()
    index = get_index(index_url)

    if index_url == DEFAULT_INDEX_URL:
        logger.error('ROSDISTRO_INDEX_URL is set to the default (did you forget to source a workspace?)')
        exit(1)

    # load rosdistro with patched SourceRepositorySpecification class
    with patch.object(repository, 'SourceRepositorySpecification', SourceRepositorySpecificationMock):
        return get_distribution(index, distroname)


def update_folder(target_path, folder_mapping, restore_version, verbose):
    # generate rosinstall file
    config = generate_rosinstall_for_repos(folder_mapping, version_tag=False, tar=False)

    for i, item in enumerate(config):
        assert len(item) == 1
        repo_type, attributes = next(iter(item.items()))

        try:
            path = attributes['local-name']
        except AttributeError as e:
            logger.warning('Repository #%d does not provide the necessary ' 'information: %s' % (i, e))
            continue
        try:
            url = attributes['uri']
            if 'version' in attributes:
                version = attributes['version']
        except AttributeError as e:
            logger.warning("Repository '%s' does not provide the necessary " 'information: %s' % (path, e))
            continue

        client = get_vcs_client(repo_type, os.path.join(target_path, path))
        if client.detect_presence():
            # TODO: backup folder and checkout the new version
            if client.get_url() != url:
                logger.error("local vcs url is different from the distro's url:\n\tlocal:  %s\n\tdistro: %s",
                             client.get_url(), url)

            target_version = None
            if restore_version:
                logger.warn("Restoring %s repo %s to version %s", repo_type, path, version)
                target_version = version

            if not client.update(version=target_version, verbose=verbose):
                logger.error("Could not update %s repo: %s", repo_type, path)
                exit(1)
        else:
            assert client.checkout(url, version=version, verbose=verbose)


cached_view = None


def get_rosdep(key):
    installer_context = create_default_installer_context(verbose=False)

    installer, installer_keys, default_key, os_name, os_version = get_default_installer(
        installer_context=installer_context, verbose=False)

    global cached_view
    if not cached_view:
        lookup = RosdepLookup.create_from_rospkg()
        lookup.verbose = False

        view = lookup.get_rosdep_view(DEFAULT_VIEW_KEY, verbose=False)
        cached_view = view
    else:
        view = cached_view

    try:
        d = view.lookup(key)
        rule_installer, rule = d.get_rule_for_platform(os_name, os_version, installer_keys, default_key)
        return rule_installer, rule
    except KeyError:
        return False


def rosdep_install(path):
    result = _rosdep_main(['install', '--from-paths', path, '-i', '-y', '--as-root', 'pip:false'])
    logger.info('rosdep exited with code %d', result)
    return result



import os
import sys
from argparse import ArgumentParser, Action, SUPPRESS

from argcomplete import autocomplete


class VersionAction(Action):
    """
    Custom VersionAction that prints the version of this package
    """

    def __init__(self,
                 option_strings,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help="show program's version number and exit"):
        super(VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        from . import __version__
        formatter = parser._get_formatter()
        python_version = '.'.join(map(str, sys.version_info[:2]))
        formatter.add_text('%(prog)s version ' + __version__ + ' (python ' + python_version + ')')
        parser.exit(message=formatter.format_help())


def parse_args(argv):
    parser = ArgumentParser()
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('--version', action=VersionAction)

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    subparser = subparsers.add_parser('install', help='install packages')
    subparser.set_defaults(func='install')
    subparser.add_argument('pkgs', nargs='+', metavar='pkg')

    subparser = subparsers.add_parser('update', help='update all packages in the workspace to the latest version')
    subparser.set_defaults(func='update')
    subparser.add_argument('--restore-versions', action='store_true',
                           help='Whether to restore the versions to the versions as defined in the rosdistro database')

    subparser = subparsers.add_parser('status', help='check if there are modified files in the workspace')
    subparser.set_defaults(func='status')

    subparser = subparsers.add_parser('list', help='List all available packages')
    subparser.set_defaults(func='list_packages')
    subparser.add_argument('--installed', action='store_true', help='Only list installed toplevel packages')

    subparser = subparsers.add_parser('remove', help='remove packages')
    subparser.set_defaults(func='remove')
    subparser.add_argument('pkgs', nargs='+', metavar='pkg')

    # workspace commands
    subparser = subparsers.add_parser(
        'ws-create',
        help='create a new workspace',
        description='Create a new catkin workspace where the packages are managed by ros-get')
    subparser.set_defaults(func='create')
    subparser.add_argument(
        'rosdistro_index_url',
        help='Custom rosdistro database url. This can be either a web url (http://) or local file url (file://).')
    subparser.add_argument(
        'extend_path',
        metavar='extend',
        help='Explicitly extend the result-space of another catkin workspace. Example: /opt/ros/kinetic')
    subparser.add_argument(
        '--dir',
        default=os.getcwd(),
        help='The directory to create the workspace in. Defaults to the current working directory')
    subparser.add_argument(
        '--name',
        help='give a name to the workspace, if not given, the name will be inferred by the directory name')
    subparser.add_argument('--build-tool', choices=['catkin_tools', 'colcon'],
                           help='Create the workspace with a specific build tool. By default, it will try to guess the '
                                'most appropriate build tool from the workspace given.')

    subparser = subparsers.add_parser('ws-switch', help='switch to a workspace')
    subparser.set_defaults(func='switch')
    subparser.add_argument('name')

    subparser = subparsers.add_parser('ws-save', help='Saves the current workspace')
    subparser.set_defaults(func='save')
    subparser.add_argument('dir')
    subparser.add_argument(
        '--name', help='give a name to the workspace, if not given, the name will be inferred by the directory name')

    subparser = subparsers.add_parser('ws-list', help='list all saved workspaces')
    subparser.set_defaults(func='list_workspaces')

    subparser = subparsers.add_parser('ws-locate', help='prints the path to the current workspace')
    subparser.set_defaults(func='locate')

    subparser = subparsers.add_parser('ws-name', help='prints the name of the current workspace')
    subparser.set_defaults(func='name')

    subparser = subparsers.add_parser('ws-rosdistro-url', help='Prints the current rosdistro_index_url')
    subparser.set_defaults(func='rosdistro_url')

    autocomplete(parser)
    args = parser.parse_args(argv)

    # computation has to be avoided before the 'autocomplete(parser)' call

    import logging.config
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': "%(log_color)s[%(levelname)s] %(name)s: %(message)s",
            }
        },
        'handlers': {
            'stream': {
                'class': 'logging.StreamHandler',
                'formatter': 'colored',
            },
        },
        'root': {
            'handlers': ['stream'],
            'level': 'INFO',
        },
    })

    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    # remove func from the namespace
    func = args.func
    del args.func
    del args.command

    return func, args


def main():
    from . import (install, update, status, list_packages, remove, create, switch, save, list_workspaces, locate, name,
                   rosdistro_url)

    func, args = parse_args(sys.argv[1:])
    # execute the function given in 'func'
    func = {
        'install': install,
        'update': update,
        'status': status,
        'list_packages': list_packages,
        'remove': remove,
        'create': create,
        'switch': switch,
        'save': save,
        'list_workspaces': list_workspaces,
        'locate': locate,
        'name': name,
        'rosdistro_url': rosdistro_url,
    }[func]
    return func(**vars(args))


if __name__ == '__main__':
    exit(main())
